"""
ensemble.py is a module designed for storing ensemble-based models that are
useful within the TensorFlow/Keras ecosystem.

The module contains:
    
    - The 'ModelGenerator' class: Instantiate a generator object that efficiently
      stores Keras models before training.
    
    - The 'DeepEnsemble' class: Instantiate a deep ensemble object with a
     '.fit()' method for efficient and sequential training of Keras models.
     The class also provides a '.predict()' method that separates the model
     outputs into lists.
    
    - The 'OutOfBagPredictor' class: Instantiate an ensemble predictor used
      exclusively for bagging models.
     
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: September 2023

import tensorflow as tf
from tensorflow import keras
from IPython.display import clear_output
import types
import numpy as np
import pandas as pd
from typing import Union, List

from aionx.kerasnn import base as knnbase
from aionx import base
from aionx.wrappers import timeit, downstream
from aionx.bootstrap import TimeSeriesBlockBootstrap

class ModelGenerator:
    """
    Author: Mikael Frenette
    
    DESCRIPTION
    ---------------------------------------------------------------------
    A generator class for efficiently storing Keras models. If the provided
    model to the constructor is a direct instance of a Keras model, then 
    the generator will simply generate the same instance with shuffled weights.
    ---------------------------------------------------------------------
    
    PARAMETERS
    ----------
    model        : The Keras model or a callable function that generates the
                   model.
    
    n_estimators : The number of models to generate.
    
    USAGE
    -----
    generator = ModelGenerator(model, n_estimators=100)
    for i, generated_model in enumerate(generator):
        # ... your code here
    
    ---------------------------------------------------------------------
    """

    def __init__(self, model: keras.models.Model, n_estimators: int = 100):

        self.model = model
        self.n_estimators = n_estimators
        self.current = 0

        # Check if the provided model is an instance of keras.models.Model
        self.iskerasmodel = isinstance(self.model, keras.models.Model)
        if self.iskerasmodel:
            self.original_weights = self.model.get_weights()

    def __iter__(self):
        return self

    def __next__(self):
        if self.current < self.n_estimators:
            if self.iskerasmodel:
                # Modify weights of the same model
                weights = [
                    np.random.permutation(w.flat).reshape(w.shape) for w in self.original_weights
                ]
                self.current += 1
                self.model.set_weights(weights)
                return self.model
                #return weights
            else:
                # Generate a new model using the provided callable function
                model = self.model()
                self.current += 1
                return model
        else:
            raise StopIteration
            
class DeepEnsemble:

    """
    Author: Mikael Frenette
    
    DESCRIPTION
    ---------------------------------------------------------------------------
    This class is used for creating and training an ensemble of neural networks
    within a bagging framework. It is operational only within a TensorFlow/Keras
    ecosystem.
    ---------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    n_estimators  : The number of estimators (neural networks) in the ensemble.
    
    network       : The base neural network(s) to use in the ensemble.
    
    trainer       : An optional network trainer for training the ensemble.
                    Default is None.
              
    sampler       : An optional bootstrapping sampler for generating training
                    samples. Default is None, which uses
                    TimeSeriesBlockBootstrap.
        
    block_size    : The block size for bootstrapping. Default is 8.
    
    sampling_rate : The sampling rate for bootstrapping. Default is 0.8.
    
    replace       : Whether to sample with replacement during bootstrapping.
                    Default is True.
    
    USAGE
    -----
    model = yourmodel(...)
    ensemble = BaggingNetwork(n_estimators=100, network=model)
    ensemble.fit(*args, **tfkwargs)
    ensemble.predict(*args, **tfkwargs)
    
    IMPORTANT NOTES
    ---------------
    - If a function is provided, no trainers should be used as it may raise
    an error by TensorFlow in later versions. The reason is that the trainer
    is using only one optimizer, so one could end up training multiple models
    with only one optimizer. This is a problem for those with internal states,
    such as keras.optimizers.Adam.
    
    - In the case where only a model is provided and, since most Keras models
    cannot be cloned (subclassing API), we need to copy the weights and simply
    shuffle them. During training, only one model will be used but with
    different sets of shuffled weight initialization.
    ---------------------------------------------------------------------------
    """

    def __init__(self,
                 n_estimators: int,
                 network: Union[tf.keras.models.Model,
                                List[tf.keras.models.Model]],
                 trainer: knnbase.NetworkTrainer = None,
                 sampler: base.Bootstrapper = None,
                 block_size: int = 8,
                 sampling_rate: float = 0.8,
                 replace: bool = True) -> None:

        self._n_estimators = n_estimators
        self._trainer = trainer
        self.network = network
        self.block_size = block_size
        self.sampling_rate = sampling_rate
        self.replace = replace
        
        self.model_generator = ModelGenerator(
            model = self.network,
            n_estimators=self._n_estimators
            )
        
        self._sampler = sampler
        
        # out of bag indices will be stored here
        self._oob_idx = {}
        
        # models will be stored here
        self._estimators = []

    @classmethod
    def from_function(cls,
                      n_estimators: int,
                      func: types.FunctionType,
                      sampler: base.Bootstrapper = None,
                      block_size: int = 8,
                      sampling_rate: float = 0.8,
                      replace: bool = True):
        """
        PARAMETERS
        ----------
        n_estimators  : The number of estimators (neural networks) in the
                        ensemble.
        
        func          : A Python function that returns a Keras model.
        
        sampler       : An optional bootstrapping sampler for generating
                        training samples.
                        Default is None, which uses TimeSeriesBlockBootstrap.
                   
        block_size    : The block size for bootstrapping. Default is 8.
        
        sampling_rate : sampling rate for bootstrapping. Default is 0.8.
        
        replace       : Whether to sample with replacement during bootstrapping.
                        Default is True.
            
        RETURNS:
            None
        """

        return cls(n_estimators, func, None, sampler, block_size, sampling_rate)
    
    
    @timeit
    def fit(self,
            X: Union[tuple[Union[np.ndarray, tf.Tensor], Union[np.ndarray, tf.Tensor]],
                     np.ndarray, tf.Tensor],
            y: Union[tuple[Union[np.ndarray, tf.Tensor], Union[np.ndarray, tf.Tensor]],
                     np.ndarray, tf.Tensor],
            epochs: int,
            batch_size: int = 32,
            validation_batch_size: int = 32,
            verbose: int = 1,
            **tfkwargs
            ) -> None:
        """
        PARAMETERS:
        -----------
        X      : The input data. It can be a single array/tensor or a tuple of
                 arrays/tensors.
            
        y      : The target data. It can be a single array/tensor or a tuple of
                 arrays/tensors.
            
        epochs : The number of training epochs.
        
        batch_size            : The batch size for training. Default is 32.
        
        validation_batch_size : The batch size for validation. Default is 32.
        
        verbose               : The verbosity mode (0, 1, or 2). Default is 1.
        
        **tfkwargs            : Additional Keras keyword arguments to pass to
                                the model.fit() call.
    
        RETURNS:
            None
        """
        
        if self._sampler is None:
            self._sampler = TimeSeriesBlockBootstrap(
                X, y, block_size=self.block_size,
                sampling_rate=self.sampling_rate, replace=True
            )
        
        for e, model in enumerate(self.model_generator):
            keras.backend.clear_session()
            X_train, y_train, X_val, y_val = next(self._sampler)

            if self._trainer is not None:
                self._trainer.train(
                    model,
                    X_train,
                    y_train,
                    epochs=epochs,
                    validation_data=(X_val, y_val),
                    batch_size=batch_size,
                    validation_batch_size=validation_batch_size,
                    **tfkwargs
                    )
                if verbose>1:
                    print(" ")
            else:
                model.fit(X_train, y_train,
                                      validation_data=(X_val, y_val),
                                      epochs=epochs,
                                      batch_size=batch_size,
                                      validation_batch_size=validation_batch_size,
                                      verbose=verbose,
                                      **tfkwargs)
                clear_output(wait=True)
                
            self._estimators.append(model.get_weights() if self.model_generator.iskerasmodel else model)
            self._oob_idx[f"Estimator_{e}"] = self._sampler.oob_indices[-1]
            
            
    def predict(self, X: Union[tuple[Union[np.ndarray, tf.Tensor],
                                     Union[np.ndarray, tf.Tensor]],
                              np.ndarray, tf.Tensor], **tfkwargs) -> list:
        """
        PARAMETERS
        ----------
        X        : Input data for predictions.
        
        tfkwargs : Additional keyword arguments for TensorFlow prediction.
        
        RETURNS
            List of predictions from individual estimators.
        """
        
        for e, model in enumerate(self._estimators):
            
            # make raw predictions
            if self.model_generator.iskerasmodel:
                self.network.set_weights(model)       
                output = self.network.predict(X, **tfkwargs)
            else:
                output = model.predict(X, **tfkwargs)
            
            if isinstance(output, list):
                output = tf.concat(output, axis=-1)
                n_outputs = output.shape[-1]
                
            if tf.rank(output) < 2:
                output = tf.expand_dims(output, axis=-1)
            n_outputs = output.shape[-1]
            
            if e == 0:
                predictions = [[] for _ in range(n_outputs)]
            
            if output.shape[-1] > 1:
                output = tf.split(output, num_or_size_splits=n_outputs, axis=-1)
                for k, out in enumerate(output):
                    predictions[k].append(out)         
            else:
                predictions[0].append(output)
                
        predictions = [
            tf.squeeze(tf.concat(pred, axis=-1)).numpy() for pred in predictions
            ]
                
        return predictions[0] if len(predictions)==1 else predictions
    
    @property
    def oob_indices(self):
        return self._oob_idx
    
    @property
    def estimators(self):
        return self._estimators



class OutOfBagPredictor(object):
    """
    Author : Mikael Frenette
    
    DESCRIPTION
    ---------------------------------------------------------------------------
    This class creates a blueprint of out-of-bag (oob) indices for each estimator
    and uses the call method to map each set of oob indices to its corresponding
    estimator."
    
    The function computes:
        oob_forecast = (1 / ((1 - sampling_rate) * n_bootstraps)) * sum(bootstraps)
        
    rows with all NaN values will produce a NaN in the oob_forecast defined
    above.
    ---------------------------------------------------------------------------  
        
    PARAMETERS
    ----------
    oob_idx : A dictionnary mapping Each estimator to its out-of-bag indices.
              each element in the dictionnary must contain an array-like value
              with out-of-bag indices.  
    ---------------------------------------------------------------------------
    """
    
    def __init__(self, oob_indices:dict)->None:
        self.oob_indices = oob_indices
        
    @downstream(
        np.ndarray, dtype="float32"
    )
    def __call__(self,
                 forecasts:np.ndarray,
                 sampling_rate:float=0.8
                 )->pd.DataFrame():
        
        """
        PARAMETERS
        ----------
        forecasts : the prediction from the models. the dataframe should have a
                    shape of (T, B) where T is the number of periods
                    and B the number of Bootstraps (i.e number of estimators).
                    The first column should be the first bootstrap while the last
                    should be the last bootstrap forecast.
            
        sampling_rate : The sampling rate parameter used by the sampler for
                        bootstrapping
                
        RETURNS
            full_forecasts : The out-of-bag as well as out-of-sample predictions
                             if the provided forecasts go beyond the training
                             set.
        """
        for i, (estimator, oob_idx) in enumerate(self.oob_indices.items()):
            rows_to_nan = np.where(~np.isin(np.arange(forecasts.shape[0]), oob_idx))[0]
            forecasts[rows_to_nan, i] = np.nan
        row_sums = np.nansum(forecasts, axis=1)
        fcst = (1 / ((1 - sampling_rate) * forecasts.shape[1])) * row_sums
        fcst[fcst==0] = np.nan
        return fcst
        
        
        



"""
aionx metrics.py is a file for storing metric classes.

The module contains:
    
    - The 'WindowDataset' class:
        A class for generating window pairs of (X, y). 
        
    - The 'Seq2SeqDataset' class:
        A class for generating sequence-to-sequence window pairs of (X, y)
        
    - The 'RegressionMatrix' class:
        A class for generating 2d regression-ready pairs of (X, y)
        
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: September 2023

import pandas as pd
import numpy as np
from typing import Union, List
from aionx import base

class WindowDataset(base.WindowHandler):
    
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    This callable class transforms a pandas dataframe into pairs of (Xs, ys)
    windows for time series forecasting. The shapes of features and labels are
    as follows: (batch, in_steps, dim) and (batch, out_steps, len(targets)),
    respectively. The data types are numpy arrays.
    
    Also, please note that if a sampler is provided (optional), calling the 
    class will return a set of training data (i.e, X_train, y_train) and a set
    of validation data (i.e, X_val, y_val). If no sampler is provided and
    targets is set to None, calling the class will only return regression
    inputs (i.e X_train).
    ---------------------------------------------------------------------------

    PARAMETERS
    ----------
    
    in_steps  : The number of time steps to consider for input sequences.
    
    out_steps : The number of time steps to predict for output sequences.
    
    horizon   : The time horizon for prediction.
    
    targets   : A list of column names to be used as target variables.
    
    sampler   : A sampler object that splits data between training and 
                validation. Defaults to None.
                             
    RETURNS
        None
        
    USAGE
    -----
    # without sampler
    generator = WindowDataset(in_steps=6, out_steps=1,
                              horizon=1, targets=["your_targets_name"])
    X_train, y_train = generator(your_training_data)
    
    #with sampler
    sampler = TimeSeriesBlockBootstrap(block_size=8, sampling_rate=0.8,
                                       replace=True)
    
    generator = WindowDataset(in_steps=6, out_steps=1,
                              horizon=1, targets=["your_targets_name"],
                              sampler=sampler)
    X_train, y_train, X_val, y_val = generator(your_training_data)
    
    """
    
    def __init__(self, targets:list, in_steps:int, horizon:int, out_steps:int=1,
                 sampler:base.Bootstrapper=None, **kwargs)->None:
        super().__init__(targets=targets,
                         in_steps=in_steps,
                         horizon=horizon,
                         out_steps=out_steps,
                         **kwargs)
        self.sampler = sampler
    
    def __call__(self, data:pd.DataFrame)->list:
        """
        PARAMETERS:
        -----------
        data : The input Pandas DataFrame containing time series data.
               The data must have the targets.
               
        RETURNS
            list containing the regression-ready window data.
        """
        # step1: create the windows
        windows = self.create_windows(data)
        
        if self.sampler is not None:
            # perform sampling and return sampled and out-of-bag indices
            self.train_idx, self.oob_idx = self.sampler(X = windows).generate_indices()

            # split X_train/y_train and X_val/y_val
            train_windows = [windows[i] for i in self.train_idx]
            val_windows = [windows[i] for i in self.oob_idx]
            X_train = np.stack([x for x, _ in train_windows], axis=0)
            y_train = np.stack([y for _, y in train_windows], axis=0)
            X_val = np.stack([x for x, _ in val_windows], axis=0)
            y_val = np.stack([y for _, y in val_windows], axis=0)
            
            if self.targets is None:
                return [X_train, X_val]
            else:
                return [X_train, y_train, X_val, y_val]
        
        else:
            # split X_train/y_train and X_val/y_val
            X_train = np.stack([x for x, _ in windows], axis=0)
            y_train = np.stack([y for _, y in windows], axis=0)
            
            if self.targets is None:
                return X_train
            else:
                return [X_train, y_train]



class Seq2SeqDataset(base.TimeSeriesDataset):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    This callable class transforms a pandas dataframe into pairs of (Xs, ys)
    windows for time series sequence-to-sequence forecasting.
    The shapes of features and labels are as follows: (batch, in_steps, dim)
    and (batch, in_steps, len(targets)), respectively. The data types are
    numpy arrays. The particularity of the Seq2SeqDataset is that the input
    steps must be the same as the output steps.
    
    Also, please note that if a sampler is provided (optional), calling the 
    class will return a set of training data (i.e, X_train, y_train) and a set
    of validation data (i.e, X_val, y_val). If no sampler is provided and
    targets is set to None, calling the class will only return regression
    inputs (i.e X_train).
    ---------------------------------------------------------------------------

    PARAMETERS
    ----------
    
    in_steps  : The number of time steps to consider for input sequences.
    
    out_steps : The number of time steps to predict for output sequences.
    
    horizon   : The time horizon for prediction.
    
    targets   : A list of column names to be used as target variables.
    
    sampler   : A sampler object that splits data between training and 
                validation. Defaults to None.
                             
    RETURNS
        None
        
    USAGE
    -----
    # without sampler
    generator = Seq2SeqDataset(in_steps=6, out_steps=6,
                              horizon=1, targets=["your_targets_name"])
    X_train, y_train = generator(your_training_data)
    
    #with sampler
    sampler = TimeSeriesBlockBootstrap(block_size=8, sampling_rate=0.8,
                                       replace=True)
    
    generator = WindowDataset(in_steps=6, out_steps=6,
                              horizon=1, targets=["your_targets_name"],
                              sampler=sampler)
    X_train, y_train, X_val, y_val = generator(your_training_data)
    
    RAISES
    ------
    ValueError : If the in_steps and out_steps do not match.
    """
    
    def __init__(self, targets:list, in_steps:int, horizon:int, out_steps:int=1,
                 sampler:base.Bootstrapper=None, **kwargs)->None:
        
        super().__init__(targets=targets,
                         in_steps=in_steps,
                         horizon=horizon,
                         out_steps=out_steps,
                         **kwargs)
        
        if in_steps!=out_steps:
            raise ValueError("In a sequence-to-sequence exercice, the output"+
                             "sequence must be the same length as the input." +
                             f"Input received : {in_steps},"+
                             f"output received:{out_steps}")
        
        self.sampler = sampler
        
        
    def _create_windows(self, data:pd.DataFrame)->list:

        # Get column indices
        column_indices = {name: i for i, name in enumerate(data.columns)}
        labels = [column_indices[i] for i in self.targets]
        windows = []
        
        i = 0 
        while True: # loop over first axis
            X = data.iloc[i:i + self.in_steps, :].values
            y = data.iloc[i + self.horizon : i + self.horizon + self.out_steps,
                          labels].values
            if (X.shape[0] < self.in_steps) or (y.shape[0] < self.out_steps):
                break
            else:
                windows.append((X, y)) # store pair of (x, y)
            i+=1
        return windows
    
    
    def __call__(self, data:pd.DataFrame)->list:
        """
        PARAMETERS:
        -----------
        data : The input Pandas DataFrame containing time series data.
               The data must have the targets.
               
        RETURNS
            list containing the regression-ready window data.
        """
        # step1: create the windows
        windows = self._create_windows(data)
        
        if self.sampler is not None:
            # perform sampling and return sampled and out-of-bag indices
            self.train_idx, self.oob_idx = self.sampler(X = windows).generate_indices()

            # split X_train/y_train and X_val/y_val
            train_windows = [windows[i] for i in self.train_idx]
            val_windows = [windows[i] for i in self.oob_idx]
            X_train = np.stack([x for x, _ in train_windows], axis=0)
            y_train = np.stack([y for _, y in train_windows], axis=0)
            X_val = np.stack([x for x, _ in val_windows], axis=0)
            y_val = np.stack([y for _, y in val_windows], axis=0)
            
            if self.targets is None:
                return [X_train, X_val]
            else:
                return [X_train, y_train, X_val, y_val]
        
        else:
            # split X_train/y_train and X_val/y_val
            X_train = np.stack([x for x, _ in windows], axis=0)
            y_train = np.stack([y for _, y in windows], axis=0)
                
            if self.targets is None:
                return [X_train, X_val]
            else:
                return [X_train, y_train, X_val, y_val]
            
            
            
class RegressionMatrix(base.TimeSeriesDataset):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    This callable class transforms a pandas dataframe into pairs of (Xs, ys)
    2d numpy arrays for time series forecasting. The shapes of features and
    labels are as follows: (batch, dim) and (batch, len(targets)),
    respectively.
    ---------------------------------------------------------------------------

    PARAMETERS
    ----------
    
    in_steps  : The number of time steps to consider for input sequences.
    
    horizon   : The time horizon for prediction.
    
    targets   : A list of column names to be used as target variables.
                             
    RETURNS
        None
        
    USAGE
    -----
    generator = RegressionMatrix(in_steps=6, horizon=1,
                                 targets=["your_targets_name"])
    
    X_train, y_train = generator(your_training_data)
    """
    
    def __init__(self,
                 targets:List[str],
                 in_steps:int,
                 horizon:int)->None:
        super().__init__(targets=targets, in_steps=in_steps, horizon=horizon)
        
    def __call__(self,
                 data:Union[pd.DataFrame, pd.Series],
                 return_dataframe:bool=False
                 ):
        """
        PARAMETERS:
        -----------
        data             : The input Pandas DataFrame containing time series
                           data. The data must have the targets.
               
        return_dataframe : If True, the class call returns a pandas
                           dataframe containing the regressors and the targets.
                           the targets will be in the first columns.
               
        RETURNS
            list containing the regression-ready window data. or a pandas
            DataFrame.
        """
        # iterate over columns
        var_lags = [data.loc[:, self.targets]]
        for col in data:
            # build lag for each columns
            lags = []
            for lag in range(1, self.in_steps+1):
                sublag = data[[col]].shift(lag+self.horizon-1)
                sublag.columns = [f'{col}_L{lag}']
                lags.append(sublag)
            var_lags.append(pd.concat(lags, axis=1))
        var_lags = pd.concat(var_lags, axis=1).dropna()
        
        if return_dataframe:
            return var_lags
        else:
            return var_lags.iloc[:, len(self.targets):].values, var_lags.iloc[:, :len(self.targets)].values
            



"""
aionx bootstrap.py is a file for storing data bootstrapping classes.

The module contains:

    - The 'TimeSeriesBlockBootstrap' class:
        A class for sampling blocks of time-dependent data. 
        
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: September 2023


import numpy as np
from typing import Iterable
from aionx import base

    
class TimeSeriesBlockBootstrap(base.Bootstrapper):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Perform block sampling on time-dependent data to preserve temporal
    structure. This class inherits from the Bootstrapper parent class and is
    designed exclusively for time series data.
    ---------------------------------------------------------------------------

    PARAMETERS
    ----------
    X             : A matrix-like containing regressors.

    y             : A matrix-like containing target variables. Default to None.
    
    block_size    : The size of time-dependent blocks to be bootstrapped.
                    Default to 8.

    sampling_rate : An amount of 'sampling_rate' will be bootstrapped from the
                            data. Default to 0.8.
                            
    replace       : If the True, sampling will be done with replacement.
                    Default to True.
                    
    **kwargs      : extra parameters inheriting from parent class.

    USAGE
    -----
    sampler = TimeSeriesBlockBootstrap(
            X=your_X, y=your_y, block_size=24, sampling_rate=0.8
    )
    X_train, y_train, X_val, y_val = next(sampler)

    """
    
    def __init__(self,
                 X,
                 y = None,
                 block_size:int=8,
                 sampling_rate:float=0.8,
                 replace:bool=True,
                 **kwargs)->None:
        
        super().__init__(sampling_rate=sampling_rate,
                         replace=replace,**kwargs)
        self.block_size = block_size
        self.X, self.y = self._configure(X), self._configure(y) 
        self.N = self._compute_length(self.X, self.y)
     
    def generate_indices(self):
        """
        DESCRIPTION
        ---------------------------------------------------------------------------
        Generates a subset of training and validation samples.
        ---------------------------------------------------------------------------

        RETURNS
            list containining a mapping indices for the bootstrapped and out-of-bag
            values.
        """
        self.iterations+=1
        # compute number of blocks
        vec = np.arange(0, self.N / self.block_size, 1)

        # draw random blocks
        groups = np.sort(np.random.choice(vec, size=self.N, replace=self.replace).astype(int))
        rando_vec = np.random.exponential(scale=1, size=(self.N // self.block_size) + 1)[groups]
        rando_vec = np.where(rando_vec > np.quantile(rando_vec, 1 - self.sampling_rate))
        chosen_one_plus = rando_vec

        bootstrap_idx = np.sort(chosen_one_plus).reshape(-1, )
        oob_idx = np.delete(np.arange(0, self.N, 1), bootstrap_idx)
        return bootstrap_idx, oob_idx
        
    def __iter__(self):
        return self

    def __next__(self):
        bootstrap_indices, oob_indices = self.generate_indices()
        self.state[f"{self.iterations}"] = {"bootstrap_indices":bootstrap_indices,
                                            "oob_indices":oob_indices}
        
        if self.X is not None:
            X_train = tuple([np.array(X)[bootstrap_indices] for X in self.X])            
            X_val = tuple([np.array(X)[oob_indices] for X in self.X])
        else:
            X_train, X_val = (None,), (None,)
            
        if self.y is not None:
            y_train = tuple([np.array(y)[bootstrap_indices] for y in self.y])      
            y_val = tuple([np.array(y)[oob_indices] for y in self.y])
        else:
            y_train, y_val = (None,), (None,)
            
        return [X_train[0] if len(X_train) == 1 else X_train,
                y_train[0] if len(y_train) == 1 else y_train,
                X_val[0] if len(X_val) == 1 else X_val,
                y_val[0] if len(y_val) == 1 else y_val]
    
    @property
    def bootstrap_indices(self):
        return self._state_history("bootstrap_indices")

    @property
    def oob_indices(self):
        return self._state_history("oob_indices")





"""
aionx scalers.py is a file for storing metric classes.

The module contains:
    
    - The 'MissingStatsError' class:
        An Exception class inheriting from base Exception.
        
    - The 'Scaler' class:
        A base class for Scaling / unscaling pandas data.
        
    - The 'StandardScaler' class:
        A class for scaling/unscaling dataframes values.

"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: August 2023

import pandas as pd
from aionx.base import Scaler
from aionx.wrappers import downstream

class MissingStatsError(Exception):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    An exception raised when summary statistics are missing.
    ---------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    message : The error message to be displayed.
    """
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class StandardScaler(Scaler):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    This class can perform Standard scaling as well as unscaling operations
    on pandas DataFrame object. 
    
    X_hat = (X - X_mean) / X_std
    ---------------------------------------------------------------------------

    USAGE
    -----
    # Create a StandardScaler instance
    scaler = StandardScaler()

    # Calculate and store statistics for a DataFrame
    data = pd.DataFrame({'feature1': [1, 2, 3, 4, 5],
                         'feature2': [2, 3, 4, 5, 6]})
    scaler.get_stats(data)

    # Scale the data
    scaled_data = scaler.scale(data)

    # Unscale the data
    feature1_unscaled = scaler.unscale(scaled_data, target='feature1')
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    
    @downstream(
        pd.DataFrame, dtype = "float32"
    )
    def scale(self, data: pd.DataFrame, target: str = None) -> pd.DataFrame:
        """
        PARAMETERS
        ----------
        data   : The input data to be scaled.
        
        target : The target variable name for which scaling should be applied.
                 If None, scaling is applied to all columns.

        RETURNS
            pd.DataFrame: The scaled data.
        """
        
        if isinstance(data, pd.Series):
            data = data.to_frame(data.name)
        
        df = data.copy()

        # Check if summary stats are available
        if not self._stats:
            raise MissingStatsError(
                " ".join([
                    "No summary stats found.",
        "Call the get_stats() method before scaling or unscaling the data."])
            )

        if target is not None:
            for col in df:
                df[col] = (df[col] - self.stats["mean"].loc[
                    target]) / self.stats["std"].loc[target]
        else:
            for col in df:
                df[col] = (df[col] - self.stats[col][
                    "mean"]) / self.stats[col]["std"]

        return df
    
    @downstream(
        pd.DataFrame, dtype="float32"
    )
    def unscale(self,
                data: pd.DataFrame,
                target: str = None,
                use_mean: bool = True) -> pd.DataFrame:
        """
        PARAMETERS
        ----------
        data     : The scaled data to be unscaled.
        
        target   : The target variable name for which unscaling should be
                   applied. If None, unscaling is applied to all columns
                   with respect to their original stats.
        
        use_mean : Whether to add the mean back during unscaling
                  (default is True).

        RETURNS
            pd.DataFrame: The unscaled data.
        """
        if isinstance(data, pd.Series):
            data = data.to_frame(data.name)
        
        df = data.copy()

        # Check if summary stats are available
        if not self._stats:
            raise MissingStatsError(
                " ".join([
                    "No summary stats found.",
        "Call the get_stats() method before scaling or unscaling the data."])
            )       
        if target is not None:
            if use_mean:
                for col in df:
                    df[col] = (df[col] * self.stats[
                        target]["std"]) + self.stats[target]["mean"]
            else:
                for col in df:
                    df[col] = df[col] * self.stats[target]["std"]
        else:
            if use_mean:
                for col in df:
                    df[col] = (df[col] * self.stats[
                        col]["std"]) + self.stats[col]["mean"]
            else:
                for col in df:
                    df[col] = df[col] * self.stats[col]["std"]
                    
        return df







"""
aionx utils.py is a file for storing/dumping utility classes.

The module contains:

    - The 'add_trends' function:
        A function for adding trends to the provided dataframe.
        
    - The 'WindowEstimationHandler' class:
        generates expanding windows from the provided pandas DataFrame.

"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: August 2023


import pandas as pd
import numpy as np
from datetime import datetime
from typing import Union
from aionx.wrappers import downstream

@downstream(
    pd.DataFrame, dtype="float32"
)
def add_trends(data: pd.DataFrame, trends: int) -> pd.DataFrame:
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Add trend columns to a given pandas DataFrame.
    ---------------------------------------------------------------------------

    PARAMETERS
    ----------
    data   : The original dataframe.

    trends : The number of trend columns to add.

    RETURNS
    -------
        The dataframe with added trend columns.
    """

    try:
        # Create trend columns as separate dataframes with values ranging
        #from 0 to len(data)
        ntrends = [pd.DataFrame(np.arange(0, len(data), 1),
                                columns=[f"trend_{i}"]) for i in range(trends)]

        # Concatenate the trend columns horizontally
        ntrends = pd.concat(ntrends, axis=1)

        # Set the index of ntrends to match the index of the original dataset
        ntrends.index = data.index

        # Concatenate the original dataset and the trend columns vertically
        data = pd.concat([data, ntrends], axis=1)
    except Exception as e:
        print(e)
    
    return data



class WindowEstimationHandler(object):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    This class is designed for managing the estimation windows in time series
    data. it generates expanding windows for the training as well as providing
    the remaining out-of-sample data.
    ---------------------------------------------------------------------------

    PARAMETERS
    ----------
    data            : The input time series data.
    
    expanding_start : The index where out-of-sample (OOS) data begins.
    
    timesteps       : The number of time steps in each window.
    
    last_window     : The index of the last window to consider.

    EXAMPLE
    -------
    # Initialize a WindowEstimationHandler
    expanding_window = WindowEstimationHandler(your_data,
                                        expanding_start=2007-01-01,
                                        timesteps = 12,
                                        last_window="2020-01-01")
    for train, oos in expanding_window:
        ...
        your code here
        ...
        
    """
    def __init__(self, data:pd.DataFrame,
                 expanding_start:Union[str, datetime],
                 timesteps:int,
                 last_window=Union[str, datetime]
                 )->None:
        
        if isinstance(last_window, str):
            last_window = datetime.strptime(last_window, "%Y-%m-%d")

        # protected attributes
        self.data = data
        self.timesteps = timesteps
        self.last_window = last_window

        # private attribute
        self.expanding_start = expanding_start

        # target values
        self.estimation_blocs = []

        # compute the number of expansions that will occur
        n_training_periods = len(self.data.loc[:self.expanding_start])
        n_oos_periods = len(self.data.loc[self.expanding_start:])
        n_expansions = int(np.ceil(n_oos_periods / timesteps))

        for expansion in range(n_expansions):
            training_end_idx = n_training_periods + (
                self.timesteps * expansion)
            forecast_end_idx = n_training_periods + (
                self.timesteps * (expansion + 1))

            # Check if last_window is reached
            if self.last_window is not None:
                if self.data.iloc[training_end_idx, :].name > self.last_window:
                    break
                else:
                    self.estimation_blocs.append(
                        (
                            self.data.iloc[:training_end_idx],
                            self.data.iloc[training_end_idx:]
                        )
                    )
            else:
                self._estimation_blocs.append(
                    (
                        self.data.iloc[:training_end_idx],
                        self.data.iloc[training_end_idx:]
                    )
                )

    def __getitem__(self, idx):
        return self.estimation_blocs[idx]

    def __len__(self):
        return len(self.estimation_blocs)



"""
aionx base.py is a file for storing base classes exclusively used in the
in the entire prototyping ecosystem.

The module contains:

    - The 'TimeSeriesDataset' class:
        parent class for time series data intput formatting. 
        
    - The 'WindowHandler' class:
        inherits from TimeSeriesDataset. parent class for window base formatting 
        child class.
        
    - The 'Bootstrapper' class:
        parent class for bootstrapping classes.
        
    - The 'Metric' class:
        parent class for metrics computation.
        
    - The 'RegressionMetric' class:
        parent class for computing regression evaluation metrics. Inherits from
        Metric
        
    - The 'DensityMetric class':
        Parent class for computin density evaluation metrics Inherits from
        Metric.
        
    - The 'Scaler' class:
        A base class for Scaling / unscaling pandas data.
"""

# Author: Mikael Frenette (mik.frenette@gmail.com)
# Created: September 2023

import numpy as np 
import pandas as pd
from typing import Union, List

class TimeSeriesDataset:
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Time Series Dataset base class.
    ---------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    targets  : The name of the target(s). Must be column's name(s).
    
    in_steps : the number of lags.
    
    horizon  : the horizon in the future to forecast.
    
    out_steps: number of timesteps to forecast at the same time.
    
    
    """
    def __init__(self,
                 targets:List[str],
                 in_steps:int,
                 horizon:int,
                 out_steps:int=1)->None:
        """
        initialize
        """
        self._targets  = targets
        self._in_steps    = in_steps
        self._horizon = horizon
        self._out_steps   = out_steps
        
    def __repr__(self):
        return '\n'.join([
            f'Total window size: {self._in_steps + self._out_steps}',
            f'Input indices: {list(np.arange(0, self._in_steps))}',
            f'Label indices: {list(np.arange(self._in_steps+self._horizon-1, self._in_steps+self._horizon+self._out_steps-1))}',
            f'Label column name(s): {self._targets}'])
    
    @property
    def targets(self):
        return self._targets
    @property
    def in_steps(self):
        return self._in_steps
    @property
    def horizon(self):
        return self._horizon
    @property
    def out_steps(self):
        return self._out_steps
    
    @targets.setter
    def targets(self, value):
        self._targets = value
    @in_steps.setter
    def in_steps(self, value):
        self._in_steps = value
    @horizon.setter
    def horizon(self, value):
        self._horizon = value
    @out_steps.setter
    def out_steps(self, value):
        self._out_steps = value



class WindowHandler(TimeSeriesDataset):
    """
    Window construction base class.
    
    PARAMETERS:
    -----------
    targets  : list of str. The name of the target(s).
                    Must be column's name(s).
                    
    in_steps : the number of lags
    
    horizon  : the horizon in the future to forecast
    
    out_steps: number of timesteps to forecast at the same time.     
    
    """
    def __init__(self,
                 targets:List[str],
                 in_steps:int,
                 horizon:int,
                 out_steps:int=1,
                 **kwargs):
        
        super().__init__(targets=targets,
                         in_steps=in_steps,
                         horizon=horizon,
                         out_steps=out_steps,
                         **kwargs)
        
    def create_windows(self, data:Union[pd.DataFrame, pd.Series])->list:
        """
        PARAMETERS
        ----------
        data : data to format into windows.

        RETURNS
            list of windows. contains tuples of pairing (X, y)

        """
        if isinstance(data, pd.Series):
            data = data.to_frame(data.name)
            
        if self.targets is None:
            self.targets = data.columns
            
        # Get column indices
        column_indices = {name: i for i, name in enumerate(data.columns)}
        labels = [column_indices[i] for i in self.targets]
        windows = []
        
        i = 0 
        while True: # loop over first axis
            X = data.iloc[i:i + self.in_steps, :].values
            y = data.iloc[
                i + self.in_steps + self.horizon - 1:i + self.in_steps +
                self.out_steps + self.horizon - 1, labels].values
            if (X.shape[0] < self.in_steps) or (y.shape[0] < self.out_steps):
                break
            else:
                windows.append((X, y)) # store pair of (x, y)
            i+=1
        return windows
    

    
class Bootstrapper(object):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Bootstrapping Parent class. Classes inheriting from Bootstrapper will gain
    2 default parameters: sampling rate and sampling with replacement.
    
    If the sampling rate is below 0.0 and higher than 1.0, an exception will be
    raised.
    ---------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    sampling_rate: An amount of 'sampling_rate' will be bootstrapped from the
                            data. Default to 0.8.
                            
    replace      : If the True, sampling will be done with replacement.
                   Default to True.
        
    RETURNS
        None
    """

    def __init__(self, sampling_rate: float=0.8,
                 replace:bool=False)->None:
        
        if not (0.0 < sampling_rate < 1.0):
            raise ValueError("sampling_rate value must be a float >0.0 and <1.0")
        else:
            self._sampling_rate = sampling_rate
        self._replace = replace
        
        self.iterations = 0
        self.reset_state() 
        
    def reset_state(self):
        """
        DESCRIPTION
        -------------------------------------------------------------------------
        resets internal history
        -------------------------------------------------------------------------
        """
        self._state = {}
        
    def _configure(self, inputs):
        if inputs is None:
            return None
        
        else:
            return (inputs, ) if not isinstance(inputs, tuple) else inputs

    def _compute_length(self, X, y):
        if X is not None:
            return len(X[0])
        else:
            if y is not None:
                return len(y[0])
            else:
                raise ValueError("Input for X or y or both must be provided. Found None.")
        
    def _state_history(self, subset):
        history = []
        for it, values in self.state.items():
            for key, vals in values.items():
                if key == subset:
                    history.append(vals)
        return history
        
    @property
    def sampling_rate(self):
        return self._sampling_rate
    
    @property
    def replace(self):
        return self._replace

    @property
    def state(self):
        return self._state
  
    
class Metric(object):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Parent class for metrics computation
    ---------------------------------------------------------------------------
    
    PARAMETERS
    ----------
    precision  : The data type that will be outputed from child's __call__
                method. One of { float32,  float64 }
    
    tolerance : The numerical tolerance parameter. this value will be used
                to control for potential underflow issues coming from the
                computations
                    
    RETURNS
        None
    
    RAISES
    ------
    ValueError : if the input for the precision parameter is invalid. 
    """
    def __init__(self, precision:str="float32",
                       tolerance:float=1e-7,
                       name:str=None)->None:
        self.precision = precision
        self.tolerance = tolerance
        self.name = name
        if self.precision not in ["float32", "float64"]:
            raise ValueError("Input for precision received {precision}. Valid" +
                             "inputs are float32 or float64")
    def _configure(self, y_true:np.ndarray, y_pred:np.ndarray)->tuple:
        if not isinstance(y_true, np.ndarray):
            y_true = np.squeeze(np.array(y_true))
        if not isinstance(y_pred, np.ndarray):
            y_pred = np.squeeze(np.array(y_pred))   
        return y_true, y_pred


class RegressionMetric(Metric):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Parent class for Regression metrics. Inherits from Metric.
    ---------------------------------------------------------------------------
    
    RAISES
    ------
    ValueError : if y_true and y_pred do not have the same shape
    """
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    def __call__(self, y_true:np.ndarray, y_pred:np.ndarray):
        """
        PARAMETERS
        ----------
        y_true : True values.
        
        y_pred : Predicted value

        RETURNS
            tuple (y_true, y_hat)
        """
        y_true, y_pred = self._configure(y_true, y_pred)
        if y_true.shape != y_pred.shape:
            raise ValueError("Input arrays must have the same shape.")
            
            
class DensityMetric(Metric):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    Parent class for Density metrics. Inherits from Metric.
    ---------------------------------------------------------------------------
    
    RAISES
    ------
    ValueError : if y_true and y_pred do not have the same shape
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    def __call__(self, y_true:np.ndarray, y_pred:Union[List[np.ndarray],
                                                       np.ndarray]):
        """
        PARAMETERS
        ----------
        y_true : True values.
        
        y_pred : Predicted value

        RETURNS
            tuple (y_true, y_hat)
        """
        y_true, y_pred = self._configure(y_true, y_pred)
        y_pred = np.split(y_pred,
                          indices_or_sections=y_pred.shape[-1], axis=1)    
        if y_true.shape != y_pred[0].shape !=y_pred[1].shape:
            raise ValueError("Input arrays must have the same shape.")
        return y_true, y_pred
    
    
    
class Scaler(object):
    """
    DESCRIPTION
    ---------------------------------------------------------------------------
    The  base class for Scalers. child inheriting from this class will have a
    get_stats method which will be able to gather information about the
    summary statistics of a provided pandas DataFrame.
    ---------------------------------------------------------------------------
    """

    def __init__(self) -> None:
        self._stats = {}
        
    def get_stats(self, data: pd.DataFrame) -> None:
        """
        Calculate and store summary statistics for data columns.

        PARAMETERS
        ----------
        data : The input data for which to calculate statistics.

        RETURNS
            None
        """
        
        if isinstance(data, pd.Series):
            data = data.to_frame(data.name)
        
        for col in data:
            summarystat = {}
            summarystat["mean"] = data[col].mean()
            summarystat["std"] = data[col].std() if data[col].std() != 0 else 1.0
            self._stats[col] = summarystat
        
    @property
    def stats(self) -> dict:
        return self._stats
    
